@@ -1,4 +1,4 @@
-#!/usr/bin/python2
+#!/usr/bin/python3
 
 import math
 import os
@@ -232,16 +232,6 @@ def walk_directories(dirs, filter_func, return_directories=False):
         #logging.critical("Error parsing directories", exc_info=True)
     return valid
 
-def rec_mkdir(path):
-    if os.path.exists(path):
-        return
-
-    rec_mkdir(os.path.split(path)[0])
-
-    if os.path.exists(path):
-        return
-    os.mkdir(path)
-
 class Section(Gtk.Box):
     def __init__(self, name):
         self.name = name
@@ -348,9 +338,10 @@ class SettingsPage(Gtk.Box):
 
         return section
 
-    def add_reveal_section(self, title, schema=None, key=None, values=None):
+    def add_reveal_section(self, title, schema=None, key=None, values=None, revealer=None):
         section = SettingsBox(title)
-        revealer = SettingsRevealer(schema, key, values)
+        if revealer is None:
+            revealer = SettingsRevealer(schema, key, values)
         revealer.add(section)
         section._revealer = revealer
         self.pack_start(revealer, False, False, 0)
@@ -380,21 +371,6 @@ class SettingsBox(Gtk.Frame):
         toolbar.add(title_holder)
         self.box.add(toolbar)
 
-        toolbar_separator = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
-        self.box.add(toolbar_separator)
-        separator_context = toolbar_separator.get_style_context()
-        frame_color = frame_style.get_border_color(Gtk.StateFlags.NORMAL).to_string()
-        css_provider = Gtk.CssProvider()
-        css_data = ".separator { -GtkWidget-wide-separators: 0; \
-                                   color: %s;                    \
-                               }" % frame_color
-        try:
-            css_provider.load_from_data(css_data)
-        except:
-            # we must be using python 3
-            css_provider.load_from_data(str.encode(css_data))
-        separator_context.add_provider(css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
-
         self.need_separator = False
 
     def add_row(self, widget):
@@ -403,7 +379,7 @@ class SettingsBox(Gtk.Frame):
             vbox.add(Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL))
         list_box = Gtk.ListBox()
         list_box.set_selection_mode(Gtk.SelectionMode.NONE)
-        row = Gtk.ListBoxRow()
+        row = Gtk.ListBoxRow(can_focus=False)
         row.add(widget)
         if isinstance(widget, Switch):
             list_box.connect("row-activated", widget.clicked)
@@ -413,19 +389,20 @@ class SettingsBox(Gtk.Frame):
 
         self.need_separator = True
 
-    def add_reveal_row(self, widget, schema=None, key=None, values=None, check_func=None):
+    def add_reveal_row(self, widget, schema=None, key=None, values=None, check_func=None, revealer=None):
         vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
         if self.need_separator:
             vbox.add(Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL))
         list_box = Gtk.ListBox()
         list_box.set_selection_mode(Gtk.SelectionMode.NONE)
-        row = Gtk.ListBoxRow()
+        row = Gtk.ListBoxRow(can_focus=False)
         row.add(widget)
         if isinstance(widget, Switch):
             list_box.connect("row-activated", widget.clicked)
         list_box.add(row)
         vbox.add(list_box)
-        revealer = SettingsRevealer(schema, key, values, check_func)
+        if revealer is None:
+            revealer = SettingsRevealer(schema, key, values, check_func)
         widget.revealer = revealer
         revealer.add(vbox)
         self.box.add(revealer)
@@ -511,7 +488,8 @@ class Switch(SettingsWidget):
         self.set_tooltip_text(tooltip)
 
     def clicked(self, *args):
-        self.content_widget.set_active(not self.content_widget.get_active())
+        if self.is_sensitive():
+            self.content_widget.set_active(not self.content_widget.get_active())
 
 class SpinButton(SettingsWidget):
     bind_prop = "value"
@@ -749,7 +727,7 @@ class Range(SettingsWidget):
 class ComboBox(SettingsWidget):
     bind_dir = None
 
-    def __init__(self, label, options=[], valtype="string", size_group=None, dep_key=None, tooltip=""):
+    def __init__(self, label, options=[], valtype=None, size_group=None, dep_key=None, tooltip=""):
         super(ComboBox, self).__init__(dep_key=dep_key)
 
         self.valtype = valtype
@@ -792,8 +770,11 @@ class ComboBox(SettingsWidget):
         self.content_widget.connect('changed', self.on_my_value_changed)
 
     def set_options(self, options):
-        # assume all keys are the same type (mixing types is going to cause an error somewhere)
-        var_type = type(options[0][0])
+        if self.valtype is not None:
+            var_type = self.valtype
+        else:
+            # assume all keys are the same type (mixing types is going to cause an error somewhere)
+            var_type = type(options[0][0])
         self.model = Gtk.ListStore(var_type, str)
 
         for option in options:
@@ -1188,5 +1169,16 @@ class Text(SettingsWidget):
         super(Text, self).__init__()
         self.label = label
 
-        self.content_widget = Gtk.Label(label=label, halign=align)
+        if align == Gtk.Align.END:
+            xalign = 1.0
+            justification = Gtk.Justification.RIGHT
+        elif align == Gtk.Align.CENTER:
+            xalign = 0.5
+            justification = Gtk.Justification.CENTER
+        else: # START and FILL align left
+            xalign = 0
+            justification = Gtk.Justification.LEFT
+
+        self.content_widget = Gtk.Label(label, halign=align, xalign=xalign, justify=justification)
+        self.content_widget.set_line_wrap(True)
         self.pack_start(self.content_widget, True, True, 0)
